package managers;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import javax.swing.JPanel;
import javax.swing.border.LineBorder;

import behaviours.BlockBehaviour;
import behaviours.NotifySpringBehaviour;
import enums.Phases;
import enums.Types;
import estructures.MapPosition;
import agents.*;
import jade.core.Agent;
import jade.core.behaviours.Behaviour;
import jade.domain.AMSService;
import jade.domain.FIPAAgentManagement.AMSAgentDescription;
import jade.domain.FIPAAgentManagement.SearchConstraints;
import jade.lang.acl.ACLMessage;
import jade.wrapper.AgentController;
import jade.wrapper.ContainerController;
import jade.wrapper.ControllerException;
import jade.wrapper.StaleProxyException;
import gui.MapGUI;

public final class MapManager {

	private static int mapWidth = 30;
	private static int mapHeight = 30;

	private static MapGUI mapGUI;
	private static LivingAgent[][] agents = new LivingAgent[mapWidth][mapHeight];
	private static HashMap<String, JPanel> agentsIcons = new HashMap<String, JPanel>();

	private static boolean initialized = false;
	private static boolean simulating = false;

	private static int plants = 0;
	private static int herbivores = 0;
	private static int carnivores = 0;

	private static Agent mapAgent;

	private MapManager() {

	}

	public static void initialize(Agent a) {
		if (!initialized) {
			mapGUI = new MapGUI();
			mapGUI.setVisible(true);
			mapAgent = a;
			initialized = true;
		}
	}

	public static boolean isInitialized() {
		return initialized;
	}

	public static void startSimulation() {
		simulating = true;
		ACLMessage msg = new ACLMessage(ACLMessage.INFORM);
		msg.addReceiver(mapAgent.getAID());
		msg.setContent("Start");

		try {
			ContainerController cc = mapAgent.getContainerController();
			AgentController ac = cc.createNewAgent("Sender",
					"agents.SenderAgent", new Object[] { msg });
			ac.start();

		} catch (ControllerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public static void resetSimulation() {
		simulating = false;
		int i = 0;
		int j = 0;
		while (i < mapWidth) {
			while (j < mapHeight) {
				if (agents[i][j] != null) {
					LivingAgent a = agents[i][j];
					// removeFromMap(a);
					a.doDelete();
					System.out.println(a.getLocalName() + ": Deleted in ("
							+ a.getMapPosition().getX() + ", "
							+ a.getMapPosition().getY() + ")");
				}
				j++;
			}
			i++;
			j = 0;
		}
		mapAgent.addBehaviour(new BlockBehaviour(mapAgent));
	}

	/**
	 * Gets all agents from the map that matches type and phase. Use null values
	 * to match all.
	 * 
	 * @param type
	 * @param phase
	 * @return
	 */
	public static List<LivingAgent> getAllAgents(Types type, Phases phase) {
		List<LivingAgent> list = new ArrayList<LivingAgent>();
		try{
		for (int i = 0; i < mapWidth; i++) {
			for (int j = 0; j < mapHeight; j++) {
				if (agents[i][j] != null
						&& (type == null || agents[i][j].getType() == type)
						&& (phase == null || agents[i][j].getPhase() == phase)) {
					list.add(agents[i][j]);

				}
			}
		}
		}catch (Exception e){}

		return list;

	}

	/**
	 * Gets all agents from the map that matches type and phase and are not
	 * furder than a distance from a position. Use null values to match all.
	 * 
	 * @param type
	 * @param phase
	 * @return
	 */
	public static List<LivingAgent> getNearbyAgents(Types type, Phases phase,
			MapPosition from, int maxDistance) {
		List<LivingAgent> list = new ArrayList<LivingAgent>();
		try{
		
		int distance=1;
		while (distance<=maxDistance && list.isEmpty()){
		for (int i = from.getX() - distance; i <= from.getX() + distance; i++) {
			for (int j = from.getY() - distance; j < from.getY() + distance; j++) {
				if (i >= 0 && i < mapWidth && j >= 0 && j < mapHeight
						&& agents[i][j] != null
						&& (type == null || agents[i][j].getType() == type)
						&& (phase == null || agents[i][j].getPhase() == phase)
						&& (from.getX()!=i || from.getY()!= j)) {
					list.add(agents[i][j]);

				}
			}
		}
		distance++;
		}
		}catch (Exception e){
			
		}

		return list;

	}

	/**
	 * Creates a LivingAgent
	 * 
	 * @param type
	 */
	public static boolean createLivingAgent(Types type, LivingAgent parent) {

		MapPosition pos = null;
		if (parent == null) {
			pos = getRandomFreePosition();
		} else {
			pos = getRandomNearbyFreePosition(parent.getMapPosition(),
					parent.getBorningDistance());
		}

		if (pos == null)
			return false;

		String name = "";
		String agent = "";
		switch (type) {
		case CARNIVORE:
			agent = "agents.CarnivoreAgent";
			name = "Carnivore" + IdManager.getCarnivoreId();
			break;
		case HERBIVORE:
			agent = "agents.HerbivoreAgent";
			name = "Herbivore" + IdManager.getHerbivoreId();
			break;
		default:
			agent = "agents.PlantAgent";
			name = "Plant" + IdManager.getPlantId();
			break;
		}

		ContainerController cc = mapAgent.getContainerController();
		AgentController ac = null;
		try {
			if (!simulating)
				return false;
			ac = cc.createNewAgent(name, agent, new Object[] { pos });

			ac.start();

		} catch (StaleProxyException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}

		return true;
	}

	public static boolean isAdyacent(MapPosition position, MapPosition to) {

		for (int i = -1; i <= 1; i++) {
			for (int j = -1; j <= 1; j++) {
				if (position.getX() + i == to.getX()
						&& position.getY() + j == to.getY()) {
					return true;
				}
			}

		}

		return false;
	}

	/**
	 * Selects a random free adyacentPosition from the map
	 * 
	 * @return
	 */
	public static MapPosition getRandomNearbyFreePosition(MapPosition position,
			int distance) {

		List<MapPosition> freeSpaces = getNearbyFreePositions(position,
				distance);

		if (freeSpaces.isEmpty())
			return null;

		int index = (int) (Math.random() * freeSpaces.size());
		MapPosition pos = freeSpaces.get(index);

		return pos;
	}

	public static List<MapPosition> getNearbyFreePositions(
			MapPosition position, int distance) {

		List<MapPosition> freeSpaces = new ArrayList<MapPosition>();
		for (int i = position.getX() - distance; i <= position.getX()
				+ distance; i++) {
			for (int j = position.getY() - distance; j <= position.getY()
					+ distance; j++) {
				if (i >= 0 && i < mapWidth && j >= 0 && j < mapHeight
						&& agents[i][j] == null) {
					freeSpaces.add(new MapPosition(i, j));
				}
			}
		}

		return freeSpaces;
	}

	/**
	 * Selects a random free position from the map
	 * 
	 * @return
	 */
	public static MapPosition getRandomFreePosition() {

		List<MapPosition> freeSpaces = new ArrayList<MapPosition>();

		// Add for free spaces
		int i = 0;
		int j = 0;
		while (i < mapWidth) {
			while (j < mapHeight) {
				if (agents[i][j] == null) {
					freeSpaces.add(new MapPosition(i, j));
				}
				j++;
			}
			i++;
			j = 0;
		}
		if (freeSpaces.isEmpty())
			return null;

		int index = (int) (Math.random() * freeSpaces.size());
		MapPosition pos = freeSpaces.get(index);

		return pos;
	}

	public static MapPosition getCloserFreeAdyacentPosition(MapPosition origin,
			MapPosition target) {

		List<MapPosition> freeSpaces = new ArrayList<MapPosition>();

		for (int i = -1; i <= 1; i++) {
			for (int j = -1; j <= 1; j++) {
				if (origin.getX() + i >= 0
						&& origin.getX() + i < mapWidth
						&& origin.getY() + j >= 0
						&& origin.getY() + j < mapHeight
						&& Math.max(
								Math.abs(origin.getX() + i - target.getX()),
								Math.abs(origin.getY() + j - target.getY())) < Math
								.max(Math.abs(origin.getX() - target.getX()),
										Math.abs(origin.getY() - target.getY()))) {
					freeSpaces.add(new MapPosition(origin.getX() + i, origin
							.getY() + j));
				}
			}
		}

		if(freeSpaces.isEmpty()) return null;
		
		int index = (int) (Math.random() * freeSpaces.size());
		MapPosition pos = freeSpaces.get(index);

		return pos;
	}

	public static void moveInMap(LivingAgent agent, MapPosition position) {
		if (!simulating)
			return;

		if (agents[position.getX()][position.getY()] != null)
			return;

		if (agent.getPhase() == Phases.DEAD)
			return;

		JPanel p = agentsIcons.get(agent.getLocalName());
		if (p != null) {
			mapGUI.movePoint(p, position);
			agents[agent.getMapPosition().getX()][agent.getMapPosition().getY()] = null;
			agents[position.getX()][position.getY()] = agent;
			agent.getMapPosition().setX(position.getX());
			agent.getMapPosition().setY(position.getY());
		}
	}

	/**
	 * Adds a LivingAgent to the map
	 * 
	 * @param agent
	 */
	public static void addToMap(LivingAgent agent) {

		if (!simulating)
			return;

		if (agents[agent.getMapPosition().getX()][agent.getMapPosition().getY()] != null)
			return;

		JPanel p = mapGUI.drawPoint(agent.getMapPosition().getX(), agent
				.getMapPosition().getY(), getColor(agent));

		agentsIcons.put(agent.getLocalName(), p);
		agents[agent.getMapPosition().getX()][agent.getMapPosition().getY()] = agent;

		switch (agent.getType()) {
		case PLANT:
			plants++;
			mapGUI.setPlants(plants);
			break;
		case HERBIVORE:
			herbivores++;
			mapGUI.setHerbivores(herbivores);
			break;
		case CARNIVORE:
			carnivores++;
			mapGUI.setCarnivores(carnivores);
			break;
		}

		// System.out.println(agent.getLocalName() + " position: ("+
		// agent.getMapPosition().getX() + ", "+ agent.getMapPosition().getY() +
		// ")");

	}

	/**
	 * Removes a LivingAgent from the map
	 * 
	 * @param agent
	 */
	public static void removeFromMap(LivingAgent agent) {
		JPanel p = agentsIcons.get(agent.getLocalName());
		if (p != null) {
			mapGUI.removePoint(p);
			agentsIcons.remove(agent.getLocalName());

			agents[agent.getMapPosition().getX()][agent.getMapPosition().getY()] = null;

			switch (agent.getType()) {
			case PLANT:
				plants--;
				mapGUI.setPlants(plants);
				break;
			case HERBIVORE:
				herbivores--;
				mapGUI.setHerbivores(herbivores);
				break;
			case CARNIVORE:
				carnivores--;
				mapGUI.setCarnivores(carnivores);
				break;
			}
		}
	}

	/**
	 * Updates the color of a LivingAgent
	 * 
	 * @param agent
	 */
	public static void updateColor(LivingAgent agent) {
		JPanel p = agentsIcons.get(agent.getLocalName());
		if (p != null) {
			Color c = getColor(agent);
			p.setBackground(c);
			p.setBorder(new LineBorder(c.darker()));
		}
	}

	/**
	 * Returns a color depending on LivingAgent's type and phase
	 * 
	 * @param agent
	 * @return
	 */
	private static Color getColor(LivingAgent agent) {
		Color color = Color.WHITE;

		switch (agent.getType()) {
		case PLANT:
			color = Color.GREEN;
			break;
		case HERBIVORE:
			color = Color.YELLOW;
			break;
		case CARNIVORE:
			color = Color.RED;
			break;
		}

		switch (agent.getPhase()) {
		case YOUNG:
			break;
		case ADULT:
			color = color.darker();
			break;
		case DEAD:
			color = color.black;
			break;
		}

		return color;
	}

}
